@@ -178,6 +178,7 @@ class WIZNET5K:  # pylint: disable=too-many-public-methods
         self.mac_address = mac
         self.src_port = 0
         self._dns = 0
+
         # Set DHCP
         self._dhcp_client = None
         if is_dhcp:
@@ -196,6 +197,17 @@ class WIZNET5K:  # pylint: disable=too-many-public-methods
         """
         if self._debug:
             print("* Initializing DHCP")
+
+        # First, wait link status is on
+        # to avoid the code during DHCP - assert self.link_status, "Ethernet cable disconnected!"
+        start_time = time.monotonic()
+        while True:
+            if self.link_status or ((time.monotonic() - start_time) > 5):
+                break
+            time.sleep(1)
+            if self._debug:
+                print("My Link is:", self.link_status)
+
         # Return IP assigned by DHCP
         self._dhcp_client = dhcp.DHCP(
             self, self.mac_address, hostname, response_timeout, debug=self._debug
